// Copyright 2001, Jonathan Bachrach.  See file TERMS.

import java.io.*;
import java.util.*;
import antlr.*;

/**
 * Fragment representing Java identifier
 * @author  Jonathan Bachrach
 * @author  Keith Playford
 * @version 0.9, 03OCT00
 */
public class IdentifierFragment extends LeafFragment implements NameFragment {
    String name;
    static int genSymNum = 0;

    static String uniquePrefix = "__";

    public String getName() { return name; }

    public String capitalize() { return capitalize(getName()); }

    public static IdentifierFragment genSym(String name) {
	return new IdentifierFragment(uniquePrefix + name + genSymNum++);
    }

    public IdentifierFragment(String s) { 
	int ttype 
	    = LooseFragmentParser .reservedWords.lookupIdentifierTokenType(s);
	name      
	    = s; 
	token     
	    = makeToken(ttype, s); 
    }   

    public IdentifierFragment(Token t) { name = t.getText(); token = t; }   

    public IdentifierFragment(String s, Token t) { name = s; token = t; }   

    public boolean equals(IdentifierFragment x) { 
	return getName().equals(x.getName());
    }
    public boolean equals(String x) { 
	return getName().equals(x);
    }
    public String getString () { return getName(); }

    void print(PrettyStream p) { p.print("Identifier: " + name); }
    void pprint(PrettyStream p, int depth, boolean isCode) { p.print(name); }
    public String toString() { return("@" + name); }
    
    public Expansion generatePatternCode
	(IdentifierFragment nowVar, IdentifierFragment nextVar) {
	// #{ FragmentList ?nextVar = ?nowVar.matchIdentifier(?(getName())) } 
	return Template.processTemplate
	    (FragmentList.fnil
	     .fpush(Template.substituteIdentifier("FragmentList"))
	     .fpush(nextVar)
	     .fpush(Template.substituteIdentifier("="))
	     .fpush(nowVar)
	     .fpush(Template.substituteDot())
	     .fpush(Template.substituteIdentifier("matchName"))
	     .fpush(Template.processParens
		    (FragmentList.fnil
		     .fpush(new StringFragment(getName()))
		     .freverse()))
	     .fpush(Template.substituteSemicolon())
	     .freverse());
    }
    public Fragment compileTemplate(int depth) {
	return Template.processTemplate
	    (FragmentList.fnil
	     .fpush(Template.substituteIdentifier
		    ("Template.substituteIdentifier"))
	     .fpush(Template.processParens
		    (FragmentList.fnil
		     .fpush(Template.substituteAsString(name))
		     .freverse()))
	     .freverse());
    }
}
