// Copyright 2001, Jonathan Bachrach.  See file TERMS.

import java.io.*;
import java.util.*;
import antlr.*;

/**
 * A pattern/template variable 
 * @author  Jonathan Bachrach
 * @author  Keith Playford
 * @version 0.9, 03OCT00
 */
public class PatternVariableFragment extends CompoundMacroFragment {
    Token              token              = null;
    static Symbol      wildcardConstraint = new Symbol("*");
    IdentifierFragment name               = null;
    SyntaxConstraint   constraint         = null;

    public SyntaxConstraint getConstraint () { return constraint; }
    public IdentifierFragment getVariableName() { return name; }
    public String getFragmentClassName () { 
	SyntaxConstraint constraint = getConstraint();
	if (constraint == null)
	    return "Fragment";
	else
	    return constraint.getFragmentClassName();
    }

    private void init (String string, Token t) {
	int ci = string.indexOf(':');
	if (ci > 0) {
	    String name       = string.substring(1, ci);
	    String constraint = string.substring(ci + 1, string.length());
	    if (name.length() == 0)
		name = constraint;
	    this.name         = new IdentifierFragment(name, t);
	    // TODO: fix this up to handle non existent constraint
	    try {
		this.constraint = MacroExpander.lookupConstraint(constraint);
	    } catch (Exception e) {
		System.err.println
		    ("Couldn't find constraint named " + constraint + " error " + e);
		e.printStackTrace();   // so we can get stack trace		
		// TODO: recover better
		this.constraint = null;
	    }
	    if (constraint == null)
		LooseFragmentParser.trace
		    ("Constraint was null input was " + string);
	} else {
	    // TODO: check for no colon case
	    String name       = string.substring(1, string.length());
	    this.name         = new IdentifierFragment(name, t);
	    // System.out.println("String = " + string + " Name = " + name);
	}
	token = t;
	// if (t.getLine() == 0)
	//     System.out.println("Null line for " + this);
    }
    public PatternVariableFragment () {
	// TODO: shouldn't be here
    }
    public PatternVariableFragment (String string) {
	init(string, makeToken(JavaTokenTypes.PATVAR, string));
    }
    public PatternVariableFragment (String string, Token t) {
	init(string, t);
    }
    public PatternVariableFragment (Token t) {
	init(t.getText(), t);
    }
    public String getName() { return name.getName(); }

    public FragmentList getInsideFragments() { 
	return FragmentList.fnil.fpush(this);
    }
    public void setInsideFragments(FragmentList fs) { /* TODO: BOGUS */ }

    public List tokens () { return List.list(token); }

    public String toString() { 
	String cs, ns;
	if (constraint == null) {
	    cs = "";
	    ns = name.getName();
	} else {
	    cs = ":" + constraint.getName();
	    if (constraint.getName().equals(name.getName()))
		ns = "";
	    else
		ns = name.getName();
	}
	return("?" + ns + cs ); 
    }
    public FragmentList collectBoundVariables () { 
	return FragmentList.fnil.fpush(this); 
    }

    public Expansion compilePatternElement
	(FragmentList tail, IdentifierFragment nowVar, IdentifierFragment endVar) {
	// #{ ?m-code; ?m*-code } package macros;

	LooseFragmentParser.trace("Constraint = " + constraint);
	if (constraint != null && constraint.isWildcardConstraint()) {
	    // #{ ?wild = FragmentList.fnil;
	    //    ?next = ?now;
	    //    while (true) {
	    //      try {
	    //        ?mTailCode;
	    //        ?name = new SequenceFragment(?wild.freverse());
	    //        break;
	    //      } catch (SyntaxMatchFailure e) {
	    //        if (?next.isEmpty())
	    //           throw e;
	    //        else {
	    //          ?wild = ?wild.fpush(?next.fhead());
	    //          ?next = ?next.ftail();
	    //          continue;
	    //        }
	    //      }
	    //    }
	    //  }

	    IdentifierFragment nextVar = IdentifierFragment.genSym("fs");
	    IdentifierFragment wildVar = IdentifierFragment.genSym("wild");
	    Expansion mTailCode        = tail.compilePatternElements(nextVar, endVar);
	    return Template.processTemplate
		(FragmentList . fnil
		 . fpush (Template.substituteIdentifier ("FragmentList"))
		 . fpush (wildVar)
		 . fpush (Template.substituteIdentifier ("="))
		 . fpush (Template.substituteIdentifier ("FragmentList"))
		 . fpush (Template.substituteDot ())
		 . fpush (Template.substituteIdentifier ("fnil"))
		 . fpush (Template.substituteSemicolon ())
		 . fpush (Template.substituteIdentifier ("FragmentList"))
		 . fpush (nextVar)
		 . fpush (Template.substituteIdentifier ("="))
		 . fpush (nowVar)
		 . fpush (Template.substituteSemicolon ())
		 . fpush (Template.substituteIdentifier ("while"))
		 . fpush (Template.processParens
			  (FragmentList . fnil
			   . fpush (Template.substituteIdentifier ("true"))
			   . freverse ()))
		 . fpush (Template.processBraces 
			  (FragmentList . fnil
			   . fpush (Template.substituteIdentifier ("try"))
			   . fpush (Template.processBraces
				    (FragmentList . fnil
				     . fpush (mTailCode)
				     . fpush (getVariableName())
				     . fpush (Template.substituteIdentifier ("="))
				     . fpush (Template.substituteIdentifier ("new"))
				     . fpush (Template.substituteIdentifier ("SequenceFragment"))
				     . fpush (Template.processParens
					      (FragmentList . fnil
					       . fpush (wildVar)
					       . fpush (Template.substituteDot ())
					       . fpush (Template.substituteIdentifier ("freverse"))
					       . fpush (Template.processParens (FragmentList . fnil))
					       . freverse ()))
				     . fpush (Template.substituteSemicolon ())
				     . fpush (Template.substituteIdentifier ("break"))
				     . fpush (Template.substituteSemicolon ())
				     . freverse ()))
			   . fpush (Template.substituteIdentifier ("catch"))
			   . fpush (Template.processParens
				    (FragmentList . fnil
				     . fpush (Template.substituteIdentifier ("SyntaxMatchFailure"))
				     . fpush (Template.substituteIdentifier ("e"))
				     . freverse ()))
			   . fpush (Template.processBraces
				    (FragmentList . fnil
				     . fpush (Template.substituteIdentifier ("if"))
				     . fpush (Template.processParens
					      (FragmentList . fnil
					       . fpush (nextVar)
					       . fpush (Template.substituteDot ())
					       . fpush (Template.substituteIdentifier ("isEmpty"))
					       . fpush (Template.processParens
							(FragmentList . fnil . freverse ()))
					       . freverse ()))
				     . fpush (Template.substituteIdentifier ("throw"))
				     . fpush (Template.substituteIdentifier ("e"))
				     . fpush (Template.substituteSemicolon ())
				     . fpush (Template.substituteIdentifier ("else"))
				     . fpush (Template.processBraces
					      (FragmentList . fnil 
					       . fpush (wildVar)
					       . fpush (Template.substituteIdentifier ("="))
					       . fpush (wildVar)
					       . fpush (Template.substituteDot ())
					       . fpush (Template.substituteIdentifier ("fpush"))
					       . fpush (Template.processParens
							(FragmentList . fnil
							 . fpush (nextVar)
							 . fpush (Template.substituteDot ())
							 . fpush (Template.substituteIdentifier ("fhead"))
							 . fpush (Template.processParens
								  (FragmentList . fnil . freverse ()))
							 . freverse ()))
					       . fpush (Template.substituteSemicolon ())
					       . fpush (nextVar)
					       . fpush (Template.substituteIdentifier ("="))
					       . fpush (nextVar)
					       . fpush (Template.substituteDot ())
					       . fpush (Template.substituteIdentifier ("ftail"))
					       . fpush (Template.processParens
							(FragmentList . fnil . freverse ()))
					       . fpush (Template.substituteSemicolon ())
					       . fpush (Template.substituteIdentifier ("continue"))
					       . fpush (Template.substituteSemicolon ())
					       . freverse ()))
				     . freverse ()))
			   . freverse ()))
		 . freverse ());
	} else {
	    IdentifierFragment nextVar = IdentifierFragment.genSym("fs");
	    Expansion mCode            = generatePatternCode(nowVar, nextVar);
	    Expansion mTailCode        = tail.compilePatternElements(nextVar, endVar);
	    return Template.processTemplate
		(FragmentList.fnil
		 .fpush(mCode)
		 .fpush(mTailCode)
		 .freverse());
	}

    }
    public Expansion generatePatternCode 
	(IdentifierFragment nowVar, IdentifierFragment nextVar) {
	// #{ SplitList tmp = ?nowVar.?matcher();
        //    FragmentList ?nextVar = tmp.getNext();
        //    ?name = tmp.getCurrent() } 
      if (constraint == null) 
	return generateAuxPatternCode(nowVar, nextVar);
      else {
	String matcher            = "matchConstraint";
	IdentifierFragment tmpVar = IdentifierFragment.genSym("tmp");
	Expansion constructor     = null;
	if (constraint.getDoesMatchMultipleFragments())
	    constructor =
		Template.processTemplate
		(FragmentList.fnil
		 .fpush(Template.substituteIdentifier("new"))
		 .fpush(Template.substituteIdentifier
			(constraint.getFragmentClassName()))
		 .fpush(Template.processParens
			(FragmentList.fnil
			 .fpush(tmpVar)
			 .fpush(Template.substituteDot())
			 .fpush(Template.substituteIdentifier("getBefore"))
			 .fpush(Template.processParens(FragmentList.fnil))
			 .freverse()))
		 .freverse());
	else 
	    constructor
		= Template.processTemplate
		(FragmentList.fnil
		 .fpush(Template.processParens
			(FragmentList.fnil
			 .fpush(Template.substituteIdentifier
				(constraint.getFragmentClassName()))
			 .freverse()))
		 .fpush(tmpVar)
		 .fpush(Template.substituteDot())
		 .fpush(Template.substituteIdentifier("getBefore"))
		 .fpush(Template.processParens(FragmentList.fnil))
		 .fpush(Template.substituteDot())
		 .fpush(Template.substituteIdentifier("getFragmentHead"))
		 .fpush(Template.processParens(FragmentList.fnil))
		 .freverse());
	return Template.processTemplate
	    (FragmentList.fnil
	     .fpush(Template.substituteIdentifier("SplitList"))
	     .fpush(tmpVar)
	     .fpush(Template.substituteIdentifier("="))
	     .fpush(nowVar)
	     .fpush(Template.substituteDot())
	     .fpush(Template.substituteIdentifier(matcher))
	     .fpush(Template.processParens
		    (FragmentList.fnil
		     .fpush(Template.substituteAsString
			    (constraint.getName()))))
	     .fpush(Template.substituteSemicolon())
	     .fpush(getVariableName())
	     .fpush(Template.substituteIdentifier("="))
	     .fpush(constructor)
	     .fpush(Template.substituteSemicolon())
	     .fpush(Template.substituteIdentifier("FragmentList"))
	     .fpush(nextVar)
	     .fpush(Template.substituteIdentifier("="))
	     .fpush(tmpVar)
	     .fpush(Template.substituteDot())
	     .fpush(Template.substituteIdentifier("getAfter"))
	     .fpush(Template.processParens(FragmentList.fnil))
	     .fpush(Template.substituteSemicolon())
	     .freverse());
      }
    }
    private Expansion generateAuxPatternCode 
	(IdentifierFragment nowVar, IdentifierFragment nextVar) {
	// #{ SplitList tmp         = new SplitList(null, null);
        //    ?name                 = ?matcher(?nowVar, tmp);
        //    FragmentList ?nextVar = tmp.getNext() }
	IdentifierFragment tmpVar = IdentifierFragment.genSym("tmp");
	return Template.processTemplate
	    (FragmentList.fnil
	     .fpush(Template.substituteIdentifier("FragmentListRef"))
	     .fpush(tmpVar)
	     .fpush(Template.substituteIdentifier("="))
	     .fpush(Template.substituteIdentifier("new"))
	     .fpush(Template.substituteIdentifier("FragmentListRef"))
	     .fpush(Template.processParens(FragmentList.fnil))
	     .fpush(Template.substituteSemicolon())
	     .fpush(getVariableName())
	     .fpush(Template.substituteIdentifier("="))
	     .fpush(name)
	     .fpush(Template.processParens
			(FragmentList.fnil
			 .fpush(nowVar)
			 .fpush(Template.substituteComma())
			 .fpush(tmpVar)
			 .freverse()))
	     .fpush(Template.substituteSemicolon())
	     .fpush(Template.substituteIdentifier("FragmentList"))
	     .fpush(nextVar)
	     .fpush(Template.substituteIdentifier("="))
	     .fpush(tmpVar)
	     .fpush(Template.substituteDot())
	     .fpush(Template.substituteIdentifier("getValue"))
	     .fpush(Template.processParens(FragmentList.fnil))
	     .fpush(Template.substituteSemicolon())
	     .freverse());
    }
    public Fragment compileTemplate (int depth) {
	// System.out.println("?x  Depth = " + depth);
	/*
	if (depth > 0)
	    return Template.processTemplate
		(FragmentList.fnil
		 .fpush(Template.substituteIdentifier
			("Template.substitutePatternVariable"))
		 .fpush(Template.processParens
			(FragmentList.fnil
			 .fpush(Template.substituteAsString("?" + name.getName()))
			 .freverse()))
		 .freverse());
	else
	    return getVariableName();
	*/
	return compileQuestionExpression(depth, 0);
    }
    public Fragment compileQuestionExpression (int tDepth, int qDepth) {
	if (tDepth == qDepth)
	    return getVariableName();
	else {
	    Fragment guts = Template.processTemplate
		(FragmentList.fnil
		 .fpush(Template.substituteIdentifier
			("Template.substitutePatternVariable"))
		 .fpush(Template.processParens
			(FragmentList.fnil
			 .fpush(Template.substituteAsString("?" + name.getName()))
			 .freverse()))
		 .freverse());
	    return QuestionExpressionFragment
		.regenerateQuestionExpression(guts, qDepth);
	}
    }
    void pprint(PrettyStream p, int depth, boolean isCode) { 
	if (isCode)
	    p.print("?" + name.getName());
	// getVariableName().pprint(p, depth, isCode);
	else
	    p.print(name); 
    }
    void pprinte(PrettyStream p, int depth, boolean isLast, boolean isCode) { 
	pprint(p, depth, isCode);
	if (!isLast) p.print(" " /* "X[" + getClass() + "]" */);
    }
    void lprint(PrettyStream p, int depth, boolean isCode) { 
	if (isCode)
	    p.print("?" + name.getName());
	// getVariableName().pprint(p, depth, isCode);
	else
	    p.print(name); 
    }
    void lprinte(PrettyStream p, int depth, boolean isLast, boolean isCode) { 
	lprint(p, depth, isCode);
	if (!isLast) p.print(" " /* "X[" + getClass() + "]" */);
    }
}

